unit gfx_compression;
{***********************************************************************
Unit gfx_compression.PAS v1.5  0801
    (c) by Andreas Moser, amoser@amoser.de,

    Delphi version : Delphi 4, 6 (due to user responses it seem to work with Delphi 5)

    gfx_compression is part of the gfx_library collection

    You may use this sourcecode for your freewareproducts.
    You may modify this source-code for your own use.
    You may recompile this source-code for your own use.

    All functions, procedures and classes may NOT be used in commercial
    products without the permission of the author. For parts of this library
    not written by me, you have to ask for permission by their
    respective authors.

    Disclaimer of warranty: "This software is supplied as is. The author
    disclaims all warranties, expressed or implied, including, without
    limitation, the warranties of merchantability and of fitness for any
    purpose. The author assumes no liability for damages, direct or
    consequential, which may result from the use of this software."

    All brand and product names are marks or registered marks of their
    respective companies.

    Please report bugs to:
    Andreas Moser  amoser@amoser.de

    The TIFFLZW - and TIFFPackBitsalgorythm are based on source codes by
        Mike Lischke (public@lischke-online.de)

********************************************************************************}

interface
uses windows,classes,sysutils;

  procedure DecodeTIFFLZW(Source, Dest: Pointer; PackedSize, UnpackedSize: Integer);
  procedure EncodeTIFFLZW(Source, Dest: Pointer; var Count: Cardinal);
  procedure DecodeTIFFRLE(Source, Dest: Pointer; PackedSize, UnpackedSize: Integer);
  procedure EncodeTIFFRLE(Source, Dest: Pointer; var Count: Cardinal);
  procedure DecodeTIFFPackBits(Source, Dest: Pointer; PackedSize, UnpackedSize: Integer);
  procedure EncodeTIFFPackBits(Source, Dest: Pointer; var Count: Cardinal);
  procedure DecodePCXRLE(Source, Dest: Pointer; PackedSize, UnpackedSize: Integer);
  procedure EncodePCXRLE(Source, Dest: Pointer; var Count: Cardinal);

  const
    ClearCode = 256;
    EOICode = 257;
    NoLZWCode = 4096;

implementation

procedure DecodeTIFFLZW(Source, Dest: Pointer; PackedSize, UnpackedSize: Integer);
var
  i: Integer;
  Data,  Bits, Code: Cardinal;
  SourcePtr: PByte;
  CodeSize: Cardinal;
  CodeMask: Cardinal;
  FreeCode: Cardinal;
  OldCode: Cardinal;
  Prefix: array[0..4095] of Cardinal;
  Suffix,
  Stack: array [0..4095] of Byte;
  StackPointer: PByte;
  TargetPtr :PByte;
  FirstChar: Byte;


function DecodeLZW(Code: Cardinal): Boolean;

var
  InCode: Cardinal; // buffer for passed code

begin
  // handling of clear codes
  if Code = ClearCode then
  begin
    // reset of all variables
    CodeSize := 9;
    CodeMask := (1 shl CodeSize) - 1;
    FreeCode := ClearCode + 2;
    OldCode := NoLZWCode;
    Result := True;
    Exit;
  end;

  // check whether it is a valid, already registered code
  if Code > FreeCode then
    raise Exception.Create('TIF LZW: invalid opcode.');

  // handling for the first LZW code: print and keep it
  if OldCode = NoLZWCode then
  begin
    FirstChar := Suffix[Code];
    TargetPtr^ := FirstChar;
    Inc(TargetPtr);
    OldCode := Code;
    Result := True;
    Exit;
  end;

  // keep the passed LZW code
  InCode := Code;  

  // the first LZW code is always smaller than FFirstCode
  if Code = FreeCode then
  begin
    StackPointer^ := FirstChar;
    Inc(StackPointer);
    Code := OldCode;
  end;

  // loop to put decoded bytes onto the stack
  while Code > ClearCode do
  begin
    StackPointer^ := Suffix[Code];
    Inc(StackPointer);
    Code := Prefix[Code];
  end;

  // place new code into code table
  FirstChar := Suffix[Code];
  StackPointer^ := FirstChar;
  Inc(StackPointer);
  Prefix[FreeCode] := OldCode;
  Suffix[FreeCode] := FirstChar;
  if FreeCode < 4096 then Inc(FreeCode);

  // increase code size if necessary
  if FreeCode = CodeMask then
  begin
    Inc(CodeSize);
    CodeMask := (1 shl CodeSize) - 1;
  end;

  // put decoded bytes (from the stack) into the target buffer
  OldCode := InCode;
  repeat
    Dec(StackPointer);
    TargetPtr^ := StackPointer^;
    Inc(TargetPtr);
  until Cardinal(StackPointer) <= Cardinal(@Stack);

  Result := True;
end;

begin
  TargetPtr := Dest;
  SourcePtr := Source;

  FreeCode := ClearCode + 2;
  OldCode := NoLZWCode;
  CodeSize := 9;
  CodeMask := (1 shl CodeSize) - 1;

  // init code table
  for I := 0 to ClearCode - 1 do
  begin
    Prefix[I] := NoLZWCode;
    Suffix[I] := I;
  end;

  // initialize stack
  StackPointer := @Stack;

  Data := 0;
  Bits := 0;
  for I := 0 to PackedSize - 1 do
  begin
    // read code from bit stream
    Inc(Data, Cardinal(SourcePtr^) shl (24 - Bits));
    Inc(Bits, 8);
    while Bits >= CodeSize do
    begin
      // current code
      Code := (Data and ($FFFFFFFF - CodeMask)) shr (32 - CodeSize);
      // mask it
      Data := Data shl CodeSize;
      Dec(Bits, CodeSize);

      // EOICode -> decoding finished, check also for badly written codes and
      // terminate the loop as soon as the target is filled up
      if (Code = EOICode) or
         ((PChar(TargetPtr) - PChar(Dest)) >= UnpackedSize) then Exit;

      if not DecodeLZW(Code) then Break;
    end;
    Inc(SourcePtr);
  end;
end;

procedure EncodeTIFFLZW(Source, Dest: Pointer; var Count: Cardinal);
begin
  //not implemented yet
end;

procedure DecodeTIFFRLE(Source, Dest: Pointer; PackedSize, UnpackedSize: Integer);
begin
  //not implemented yet
end;

procedure EncodeTIFFRLE(Source, Dest: Pointer; var Count: Cardinal);
begin
  //not implemented yet
end;

procedure DecodeTIFFPackBits(Source, Dest: Pointer; PackedSize, UnpackedSize: Integer);
// decodes a simple run-length encoded strip of size PackedSize

var
  SourcePtr,
  TargetPtr: PByte;
  N: SmallInt;

begin
  TargetPtr := Dest;
  SourcePtr := Source;
  while PackedSize > 0 do
  begin
    N := ShortInt(SourcePtr^);
    Inc(SourcePtr);
    Dec(PackedSize);
    if N < 0 then // replicate next Byte -N + 1 times
    begin
      if N = -128 then Continue; // nop
      N := -N + 1;
      FillChar(TargetPtr^, N, SourcePtr^);
      Inc(SourcePtr);
      Inc(TargetPtr, N);
      Dec(PackedSize);
    end
    else
    begin // copy next N + 1 bytes literally
      Move(SourcePtr^, TargetPtr^, N + 1);
      Inc(TargetPtr, N + 1);
      Inc(SourcePtr, N + 1);
      Dec(PackedSize, N + 1);
    end;
  end;
end;

procedure EncodeTIFFPackBits(Source, Dest: Pointer; var Count: Cardinal);
begin
  //not implemented yet
end;

procedure DecodePCXRLE(Source, Dest: Pointer; PackedSize, UnpackedSize: Integer);
Var
  databyte,db2:byte;
  repeater:SmallInt;
  Planes,n,p:Word;
Begin
 n:=0;
 repeat

    databyte:=PByte(Source)^;
    inc(PByte(Source));
    If databyte >=$C0 Then
    Begin
      repeater:=databyte And $3F;
      db2:=PByte(Source)^;
      inc(PByte(Source));
      while repeater>0 do
      begin
        pByte(Dest)^:=db2;
        Inc(PByte(Dest));
        dec(repeater);
        Inc(n);
      end;
    end else
      Begin
        pByte(Dest)^:=dataByte;
        Inc(PByte(Dest));
        Inc(n);
      end;
 until n>=UnpackedSize ;
end;

procedure EncodePCXRLE(Source, Dest: Pointer; var Count: Cardinal);
begin
  //not implemented yet
end;


end.
